#include "exchange.h"
#ifndef timing_DEF
   #define timing_DEF
   #include "timing.h"
#endif
#include <stdlib.h>
#include <stdio.h>
#include <mpi.h>






/****************************************************************************************************/
  void initialize_MPI(int *argc, char *** argv){

     MPI_Init(argc, argv);

     comm_comm=MPI_COMM_WORLD;

     MPI_Comm_rank(comm_comm, &my_rank);
     printf ("PE%i:\t initialize_MPI: MPI_Comm_rank=%d \n", my_rank,my_rank);

     MPI_Comm_size(comm_comm, &comm_size);
     printf ("PE%i:\t initialize_MPI: MPI_Comm_size=%d \n", my_rank,comm_size);
   }
/****************************************************************************************************/
  void finish_MPI(){
      MPI_Finalize();
      printf("PE%i: has finalized MPI \n",my_rank);
      exit(0);
  }
/****************************************************************************************************/
  void exit_MPI(int error ){
      printf("PE%i: will abort with error %d \n",my_rank,error);
      fflush(NULL);
      MPI_Abort(comm_comm, error);
      exit(error);
  }
/****************************************************************************************************/
void allocate_exchange_buffer(exchange_buffer * buffer , int length , int offset, int indirect, int data_type, int tag )
 {


      buffer->tag=tag;
      buffer->offset=offset;
      buffer->indirect=indirect;
      buffer->data_type=data_type;

      buffer->length_index=0;
      buffer->length_double=0;
      buffer->length_int=0;
      buffer->index      =NULL;
      buffer->double_val =NULL;
      buffer->int_val    =NULL;

   if(length > 0 ) {
      if(indirect == 1) { 
	buffer->length_index=length;
	buffer->index=malloc(sizeof(int   )*length);
	} else {
	buffer->length_index=0;
	buffer->index      =NULL;
	}
      if(data_type == 1) { 
	buffer->length_double =length;
      	buffer->double_val  =malloc(sizeof(double)*length);
	}
      if(data_type == 2) {
	buffer->length_int =length;
	buffer->int_val     =malloc(sizeof(double)*length);
	}
       }
 }
/****************************************************************************************************/
void deallocate_exchange_buffer(exchange_buffer * buffer)
 {


      buffer->tag=0;
      buffer->offset=0;
      buffer->indirect=0;
      buffer->data_type=0;

	if( buffer->length_index  > 0 ) free(buffer->index);
      	if( buffer->length_double > 0 ) free(buffer->double_val);
	if( buffer->length_int    > 0 ) free(buffer->int_val)  ;
 }
/****************************************************************************************************/
void allocate_buffer_set(buffer_set * all_buffers , int number_of_buffers )
 {
   if(number_of_buffers <= 0 ) {
      all_buffers->number_of_buffers=0;
      all_buffers->buffers=NULL;
      all_buffers->requests=NULL;
       }
   else {
      all_buffers->number_of_buffers=number_of_buffers;
      all_buffers->buffers= malloc(sizeof(exchange_buffer)*number_of_buffers);
      all_buffers->requests=malloc(sizeof(MPI_Request    )*number_of_buffers);
       }
 }
/****************************************************************************************************/
void fill_exchange_buffer(exchange_buffer buffer, double * double_values, int * int_values) {
	 int ii;
	 int length=buffer.length_index;
         double time;
         int offset=buffer.offset;

    //     printf ("PE%i:\t initiate_send: buffer will be filled  \n", my_rank);

	if(buffer.indirect == 1 ) {
	if(buffer.data_type == 1 ) {
	 for (ii=0; ii<length; ii++) { 
		 buffer.double_val[ii]=double_values[offset+buffer.index[ii]];
		}
	} else if(buffer.data_type == 2 ){
	 for (ii=0; ii<length; ii++) { 
		 buffer.int_val[ii]=int_values[offset+buffer.index[ii]];
		}
	}

	} else {

	if(buffer.data_type == 1 ) {
	 for (ii=0; ii<length; ii++) { 
		 buffer.double_val[ii]=double_values[offset+ii];
		}
	} else if(buffer.data_type == 2 ){
	 for (ii=0; ii<length; ii++) { 
		 buffer.int_val[ii]=int_values[offset+ii];
		}
	}
	}
}
/****************************************************************************************************/
void initiate_send_buffer(double * double_values, int * int_values, exchange_buffer buffer, MPI_Request * request)
 {
   double time_start;
   //      printf ("PE%i:\t initiate_send: buffer will be sent to %i\n", my_rank, buffer.to_proc);

	fill_exchange_buffer(buffer, double_values, int_values) ;

         // data will be sent. The request handle is given back
//time_start=MPI_Wtime();
	if(buffer.data_type == 1 ) {
	 MPI_Isend(buffer.double_val, buffer.length_double, MPI_DOUBLE, buffer.to_proc, buffer.tag, comm_comm, request);
	} else if(buffer.data_type == 2 ){
	 MPI_Isend(buffer.int_val, buffer.length_int, MPI_INT, buffer.to_proc, buffer.tag, comm_comm, request);
	}
//time=MPI_Wtime()-time_start;
//   printf ("PE%i:\t  time for MPI_Isend %lf muesec \n", my_rank, time*1.e6);
 }
/****************************************************************************************************/
void initiate_send_buffer_set( double * double_values, int * int_values, buffer_set to_buffer_set)
 {
  int ii;
  for (ii=0; ii< to_buffer_set.number_of_buffers ; ii++ ) {
      initiate_send_buffer(double_values, int_values, to_buffer_set.buffers[ii], &(to_buffer_set.requests[ii]));
  }
 }
/****************************************************************************************************/
void complete_send_buffer_set(buffer_set to_buffer_set, MPI_Status * status_array)
 {

	 MPI_Waitall(to_buffer_set.number_of_buffers, to_buffer_set.requests, status_array);
// status.MPI_SOURCE // von welchem rank kommt die Nachricht 
// status.MPI_TAG    // welchen tag hat die Nachricht
// status.MPI_ERRROR // error code  MPI_SUCCESS
// MPI_Get_count(MPI_Status * status, MPI_DOUBLE, int *count);
		      
 }
/****************************************************************************************************/
void initiate_receive_buffer(exchange_buffer buffer, MPI_Request * request)
 {
	if(buffer.data_type == 1 ) {
	 MPI_Irecv(buffer.double_val, buffer.length_double, MPI_DOUBLE, buffer.from_proc, buffer.tag, comm_comm, request);
	} else if(buffer.data_type == 2 ){
	 MPI_Irecv(buffer.int_val, buffer.length_int, MPI_INT, buffer.from_proc, buffer.tag, comm_comm, request);
	}
     //    printf ("PE%i:\t initiate_receive: receive initiated for %i\n", my_rank, buffer.from_proc);
 }
/****************************************************************************************************/
void initiate_receive_buffer_set(buffer_set from_buffer_set)
 {
  int ii;
  for (ii=0; ii< from_buffer_set.number_of_buffers ; ii++ ) {
      initiate_receive_buffer(from_buffer_set.buffers[ii], &(from_buffer_set.requests[ii]));
  }
 }
/****************************************************************************************************/
void apply_exchange_buffer(exchange_buffer buffer, double * double_values, int * int_values)
 {
	 int ii;
	 int length=buffer.length_index;
         int offset=buffer.offset;

	if(buffer.indirect == 1 ) {
	if(buffer.data_type == 1 ) {
	 for (ii=0; ii<length; ii++) { 
		 double_values[offset+buffer.index[ii]]=buffer.double_val[ii];
		}
	} else if(buffer.data_type == 2 ){
	 for (ii=0; ii<length; ii++) { 
		 int_values[offset+buffer.index[ii]]=buffer.int_val[ii];
		}
	}
	} else {
	if(buffer.data_type == 1 ) {
	 for (ii=0; ii<length; ii++) { 
		 double_values[offset+ii]=buffer.double_val[ii];
		}
	} else if(buffer.data_type == 2 ){
	 for (ii=0; ii<length; ii++) { 
		 int_values[offset+ii]=buffer.int_val[ii];
		}
	}
	}
 }
/****************************************************************************************************/
void complete_receive_buffer_set(double * double_values, int * int_values, buffer_set from_buffer_set )
 {
        int incount;
        int j_finished;
        int outcount;
        int ii;
        int array_of_indices[NUMBER_OF_BUFFERS];
        MPI_Status array_of_statuses[NUMBER_OF_BUFFERS];


	   do {

        // wait for finishing
	   MPI_Waitsome(from_buffer_set.number_of_buffers
                      , from_buffer_set.requests
                      , &outcount
                      , array_of_indices
                      , array_of_statuses); 

//	   if ( outcount == MPI_UNDEFINED ) break;
  //             printf ("PE%i:\t complete_receive_buffer_set: got new data from %i processes \n", my_rank, outcount);

        // if some communications are ready we read the data
	    for ( j_finished = 0 ; j_finished < outcount ; j_finished++)  {
               apply_exchange_buffer(from_buffer_set.buffers[array_of_indices[j_finished]], double_values, int_values);
	            }
	       } while ( outcount != MPI_UNDEFINED ) ;
 }
/****************************************************************************************************/
//void complete_sent_buffer_set(double * double_values, int * int_values, int number_of_from_buffers, exchange_buffer * buffer, MPI_Request * array_of_requests)
// {
//        int incount;
//        int j_finished;
//        int outcount;
//        int ii;
//        int array_of_indices[NUMBER_OF_BUFFERS];
//        MPI_Status array_of_statuses[NUMBER_OF_BUFFERS];
//
//
//	   do {
//
//	   MPI_Waitsome(number_of_from_buffers, array_of_requests, &outcount, array_of_indices, array_of_statuses); 
//
////	   if ( outcount == MPI_UNDEFINED ) break;
//   //            printf ("PE%i:\t complete_receive_buffer_set: got new data from %i processes \n", my_rank, outcount);
//
//	    for ( j_finished = 0 ; j_finished < outcount ; j_finished++)  {
//               apply_exchange_buffer(buffer[array_of_indices[j_finished]], double_values, int_values);
//	            }
//	       } while ( outcount != MPI_UNDEFINED ) ;
// }
/****************************************************************************************************/
double dot_pro( double * values_1, double * values_2, int length )
 {
  int ii;
  int count=1;
  double sum;
  double time;
  double received_sum;
   double time_start;

  sum=0.;
  for (ii=0; ii< length ; ii++ ) {
      sum=sum+values_1[ii]*values_2[ii];
  }

   time_start=MPI_Wtime();
      MPI_Allreduce(&sum,&received_sum,count,MPI_DOUBLE,MPI_SUM,comm_comm);
   time=MPI_Wtime()-time_start;

//   printf ("PE%i:\t dot_pro: %f \n", my_rank, sum);
//   printf ("PE%i:\t dot_pro: time for MPI_Allreduce %f muesec \n", my_rank, time*1.e6);

  return(received_sum);
 }
/****************************************************************************************************/
/****************************************************************************************************/
/****************************************************************************************************/
/****************************************************************************************************/
void initiate_send_exch_set(exch_set_type exch_set)
 {
        int length=exch_set.length;
        int buf_length;
        int ll;
        MPI_Datatype datatype;
   double time_start;

//         printf ("PE%i:\t initiate_send_exch_set: data will be sent to environment \n", my_rank);

        datatype=exch_set.datatype;
//time_start=MPI_Wtime();
         for(ll=0; ll < length; ll++ ){
             buf_length=exch_set.to_buf_length[ll];
	 MPI_Isend(exch_set.to_buf[ll], buf_length, datatype, exch_set.proc[ll]
                  , PARAMETER_TAG, comm_comm, &(exch_set.to_request_array[ll]));
         }
//time=MPI_Wtime()-time_start;
//   printf ("PE%i:\t  time for MPI_Isend %lf muesec \n", my_rank, time*1.e6);
 }
/****************************************************************************************************/
void initiate_receive_exch_set(exch_set_type exch_set)
 {
        int length=exch_set.length;
        int buf_length;
        int ll;
        MPI_Datatype datatype;

        datatype=exch_set.datatype;

         for(ll=0; ll < length; ll++ ){
             buf_length=exch_set.from_buf_length[ll];
	 MPI_Irecv(exch_set.from_buf[ll], buf_length, datatype, exch_set.proc[ll]
                  , PARAMETER_TAG, comm_comm, &(exch_set.from_request_array[ll]));
           }

 }
/****************************************************************************************************/
void complete_receive_exch_set(exch_set_type exch_set)
 {
        int incount;
        int j_finished;
        int outcount;
        int ii;


	   do {

        // wait for finishing
	   MPI_Waitsome(exch_set.length
                      , exch_set.from_request_array
                      , &outcount
                      , exch_set.temp
                      , exch_set.from_status_array); 

//	   if ( outcount == MPI_UNDEFINED ) break;
//             printf ("PE%i:\t complete_receive_buffer_set: got new data from %i processes \n", my_rank, outcount);

        // if some communications are ready we read the data

//    for ( j_finished = 0 ; j_finished < outcount ; j_finished++)  {
//         printf ("PE%i:\t complete_receive_exch_set: got data for process %d \n", my_rank,exch_set.proc[exch_set.temp[j_finished]]);
//	 }
	       } while ( outcount != MPI_UNDEFINED ) ;
 }
/****************************************************************************************************/
void complete_send_exch_set(exch_set_type exch_set)
 {

	 MPI_Waitall(exch_set.length, exch_set.to_request_array, exch_set.to_status_array);
		      
 }
/****************************************************************************************************/
void sendreceive_exch_set(exch_set_type exch_set)
 {


//     printf ("PE%i:\t sendreceive_exch_set: initiate_send_exch_set \n", my_rank);
  initiate_send_exch_set(exch_set);
//     printf ("PE%i:\t sendreceive_exch_set: initiate_receive_exch_set \n", my_rank);
  initiate_receive_exch_set(exch_set);
//     printf ("PE%i:\t sendreceive_exch_set: complete_receive_exch_set \n", my_rank);
  complete_receive_exch_set(exch_set);
//     printf ("PE%i:\t sendreceive_exch_set: complete_send_exch_set \n", my_rank);
  complete_send_exch_set(exch_set);
//     printf ("PE%i:\t sendreceive_exch_set: finished \n", my_rank);
		      
 }
/****************************************************************************************************/
void allocate_exch_set(exch_set_type * exch , int length )
 {



	exch->length=0;
  	exch->datatype=MPI_INT;
	exch->temp         =NULL;
	exch->proc         =NULL;
	exch->from_request_array=NULL;
	exch->from_status_array =NULL;
	exch->from_buf_length   =NULL;
	exch->from_buf          =NULL;
	exch->to_request_array  =NULL;
	exch->to_status_array   =NULL;
	exch->to_buf_length     =NULL;
	exch->to_buf            =NULL;

   if(length > 0 ) {
	exch->length=length;
	exch->temp              =malloc(sizeof(int        )*length);
	exch->proc              =malloc(sizeof(int        )*length);
	exch->from_request_array=malloc(sizeof(MPI_Request)*length);
	exch->from_status_array =malloc(sizeof(MPI_Status )*length);
	exch->from_buf_length   =malloc(sizeof(int        )*length);
	exch->from_buf          =malloc(sizeof(void*      )*length);
	exch->to_request_array  =malloc(sizeof(MPI_Request)*length);
	exch->to_status_array   =malloc(sizeof(MPI_Status )*length);
	exch->to_buf_length     =malloc(sizeof(int        )*length);
	exch->to_buf            =malloc(sizeof(void*      )*length);
       }
 }
/****************************************************************************************************/
void define_neighbour_processors_exch_set(exch_set_type * exch ,int * array){

int ll;
int length;

length=exch->length;

for( ll=0; ll < length; ll++) {
     printf ("PE%i:\t define_neighbour_processors_exch_set: proc[%d]=%d \n", my_rank,ll,array[ll]);
   exch->proc[ll]=array[ll];
}


}
/****************************************************************************************************/
void hide_neighbour_processors_exch_set(exch_set_type * exch ,int * array){

int ll;
int length;

length=exch->length;

for( ll=0; ll < length; ll++) {
   if( array[ll] == 0 ) {
     printf ("PE%i:\t hide_neighbour_processors_exch_set: proc[%d]=%d set to %d \n", my_rank,ll,exch->proc[ll],MPI_PROC_NULL);
      exch->proc[ll]=MPI_PROC_NULL;
    }
}

}
/**************************************************************************************************/
void variables_exch_set(exch_set_type * exch, MPI_Datatype datatype,void * to_array, void *  from_array){

int ll;
int length;
MPI_Aint extent;

length=exch->length;

   exch->datatype=datatype;

   MPI_Type_extent(datatype, &extent);
 //    printf ("PE%i:\t variables_exch_set: datatype_extent=%d \n", my_rank,extent);

for( ll=0; ll < length; ll++) {
   exch->to_buf_length[ll]=1;
   exch->to_buf[ll]=to_array + extent*ll;
   exch->from_buf_length[ll]=1;
   exch->from_buf[ll]=from_array + extent*ll;
}

   sendreceive_exch_set(*exch);
}
/**************************************************************************************************/
void arrays_exch_set(exch_set_type * exch, MPI_Datatype datatype
                        ,void * to_array,int * to_offset_array, int * to_length_array
                        ,void * from_array,int * from_offset_array, int * from_length_array
                        ){
int ll;
int length;
int to_offset,from_offset;
MPI_Aint extent;

length=exch->length;

   exch->datatype=datatype;

   MPI_Type_extent(datatype, &extent);

//     printf ("PE%i:\t arrays_exch_set: datatype_extent=%d \n", my_rank,extent);

for( ll=0; ll < length; ll++) {
   exch->to_buf_length[ll]  =to_length_array[ll];
                 to_offset  =to_offset_array[ll];
   exch->to_buf[ll]         =to_array + extent*to_offset;
   exch->from_buf_length[ll]=from_length_array[ll];
               from_offset  =from_offset_array[ll];
   exch->from_buf[ll]       =from_array + extent*from_offset;
}
   sendreceive_exch_set(*exch);
}
/**************************************************************************************************/
