#ifndef patch_DEF
   #define patch_DEF
   #include "patch.h"
#endif
#ifndef sparse_matrix_DEF
   #define sparse_matrix_DEF
   #include "sparse_matrix.h"
#endif
#include <stdlib.h>
#include <limits.h>
#include <string.h>
#include "index_sort.h"

#define INFINITY INT_MAX




/**************************************************************************************/
 void test_matrix_type(sparse_matrix_type matrix)
 {
    int total_number_of_operations;
   double time_start;
   double time_end;

   time_start=MPI_Wtime();

//printf("PE%i:\t matrix.mat_number_of_rows = %d\n",my_rank,matrix.mat_number_of_rows);

           test_matrix_type_kernel(
                                   matrix.mat_aa
                                  ,matrix.mat_index
                                  ,matrix.mat_elements
                                  ,matrix.mat_number_of_row_elements
                                  ,matrix.mat_number_of_rows
                                  ,matrix.mat_number_of_entries
                                  ,&(matrix.matrix_type)
                                  ,&total_number_of_operations
                                  ) ;
   time_end=MPI_Wtime();
   printf("PE%i:\t sparse_matrix_vector: time %f muesec \n",my_rank, (time_end-time_start)*1.e6);
   printf("PE%i:\t sparse_matrix_vector: Performance %f MFLOPs \n",my_rank,
                               (double)(total_number_of_operations)/(time_end-time_start)*1.e-6);
 }
/**************************************************************************************/
   void test_matrix_type_kernel(
                                   double * mat_aa
                                  ,int * mat_index
                                  ,int * mat_elements
                                  ,int * mat_number_of_row_elements
                                  ,int mat_number_of_rows
                                  ,int mat_number_of_entries
                                  ,int *matrix_type
                                  ,int * total_number_of_operations
                                  )
{
 int ii,nn,kk,num,nn_max;
 double sum;
 int status;


 status=0;
 num=-1;
 for( ii = 0 ; ii < mat_number_of_rows; ii++ ) {
     if(mat_elements[ii] !=ii){
         status=1;
     }
 }

if(status == 1 ){
 *matrix_type = 1;
}

}

/**************************************************************************************/
 void matrix_vector( vector_type  yy, sparse_matrix_type matrix, vector_type xx)
 {
    int total_number_of_operations;
   double time_start;
   double time_end;

   time_start=MPI_Wtime();

//printf("PE%i:\t matrix.mat_number_of_rows = %d\n",my_rank,matrix.mat_number_of_rows);

           sparse_matrix_vector_kernel(
                                   yy.values
                                  ,matrix.mat_aa
                                  ,matrix.mat_index
                                  ,matrix.mat_elements
                                  ,matrix.mat_number_of_row_elements
                                  ,matrix.mat_number_of_rows
                                  ,matrix.mat_number_of_entries
                                  ,&(matrix.matrix_type)
                                  ,xx.values
                                  ,&total_number_of_operations
                                  ) ;
   time_end=MPI_Wtime();
   printf("PE%i:\t sparse_matrix_vector: time %f muesec \n",my_rank, (time_end-time_start)*1.e6);
   printf("PE%i:\t sparse_matrix_vector: Performance %f MFLOPs \n",my_rank,
                               (double)(total_number_of_operations)/(time_end-time_start)*1.e-6);
 }
/**************************************************************************************/
   void sparse_matrix_vector_kernel(
                                   double * yy
                                  ,double * mat_aa
                                  ,int * mat_index
                                  ,int * mat_elements
                                  ,int * mat_number_of_row_elements
                                  ,int mat_number_of_rows
                                  ,int mat_number_of_entries
                                  ,int *matrix_type
                                  ,double * xx
                                  ,int * total_number_of_operations
                                  )
{
 int ii,nn,kk,num,nn_max;
 double sum;


if( *matrix_type == 0 ) {
 num=-1;
 for( ii = 0 ; ii < mat_number_of_rows; ii++ ) {
  sum = 0;
  nn_max = mat_number_of_row_elements[ii];
  for( nn = 0; nn < nn_max ; nn++) {
   num = num+1;
   sum = sum+mat_aa[num]*xx[mat_index[num]];
  }
  yy[mat_elements[ii]] = sum;
 }
*total_number_of_operations=2*num;
}


else if( *matrix_type == 1 ) {

  /* case with direct adressing of elements */
 num=-1;
 for( ii = 0 ; ii < mat_number_of_rows; ii++ ) {
  yy[ii] = 0;
  nn_max = mat_number_of_row_elements[ii];
  for( nn = 0; nn < nn_max ; nn++) {
   num = num+1;
   yy[ii] = yy[ii]+mat_aa[num]*xx[mat_index[num]];
  }
}

}

}

/**************************************************************************************/
void write_sparse_matrix(char string[], sparse_matrix_type matrix) {

 printf("PE%i:\t \n",my_rank);
 printf("PE%i:\t \n",my_rank);
 printf("PE%i:\t %s\n",my_rank,string);
 printf("PE%i:\t matrix.mat_number_of_rows=%d\n",my_rank,matrix.mat_number_of_rows);
           write_sparse_matrix_kernel(
                                   matrix.mat_aa
                                  ,matrix.mat_index
                                  ,matrix.mat_elements
                                  ,matrix.mat_number_of_row_elements
                                  ,matrix.mat_number_of_rows
                                  ,matrix.mat_number_of_entries
                                  ) ;
 printf("PE%i:\t end of %s\n",my_rank,string);
   printf("PE%i:\t \n",my_rank);
   printf("PE%i:\t \n",my_rank);
}
/**************************************************************************************/
void write_sparse_matrix_kernel(
                                   double * mat_aa
                                  ,int * mat_index
                                  ,int * mat_elements
                                  ,int * mat_number_of_row_elements
                                  ,int mat_number_of_rows
                                  ,int mat_number_of_entries
                                  ) {
  int ii,nn,kk,num,nn_max,nn_start;

  printf("PE%i:\t write_sparse_matrix_kernel: \n",my_rank);
  printf("PE%i:\t    mat_number_of_rows = %d \n",my_rank,mat_number_of_rows);
  printf("PE%i:\t mat_number_of_entries = %d \n",my_rank,mat_number_of_entries);

  num=-1;
   printf("PE%i:\t \n",my_rank);
   printf("PE%i:\t number|element| indices ....\n",my_rank);
   printf("PE%i:\t ------+-------+-----------------------------------------------------------\n",my_rank);
  for( ii = 0 ; ii < mat_number_of_rows; ii++ ) {
        nn_max = mat_number_of_row_elements[ii];
        nn_start=num;
   printf("PE%i:\t %5d | %5d | ",my_rank,ii,mat_elements[ii]);
     for( nn = 0; nn < nn_max ; nn++) {
          num = num+1;
          printf("PE%i:\t %8d ",my_rank,mat_index[num]);
     }
    printf("PE%i:\t \n",my_rank);
    printf("PE%i:\t       |  ",my_rank);
    printf("PE%i:\t     |   ",my_rank);
     num=nn_start;
         for( nn = 0; nn < nn_max ; nn++) {
              num = num+1;
              printf("PE%i:\t %8.3f ",my_rank,mat_aa[num]);
         }
    printf("PE%i:\t \n",my_rank);
  }
   printf("PE%i:\t ------+-------+-----------------------------------------------------------\n",my_rank);

}
/**************************************************************************************/
void renumber_sparse_matrix(sparse_matrix_type matrix, renumber_type * local_to_global) {
  int * mapping_elements;
  int * index_reference;
  int ii ;
   double time_start;
   double time_end;

   time_start=MPI_Wtime();
 printf("PE%i:\t matrix.mat_number_of_rows=%d\n",my_rank,matrix.mat_number_of_rows);
           renumber_sparse_matrix_kernel(
                                   matrix.mat_index
                                  ,matrix.mat_elements
                                  ,matrix.mat_number_of_rows
                                  ,matrix.mat_number_of_entries
				  ,local_to_global
                                  ) ;
   time_end=MPI_Wtime();
   printf("PE%i:\t renumber_sparse_matrix: time %f muesec \n",my_rank, (time_end-time_start)*1.e6);
}
/**************************************************************************************/
void renumber_sparse_matrix_kernel(
                                   int * mat_index
                                  ,int * mat_elements
                                  ,int mat_number_of_rows
                                  ,int mat_number_of_entries
				  ,renumber_type * loc_to_global
                                  ) {
  int ii,jj;
  int length_local_to_global;
  int index;
  int element;
  int * renumber_elements;
  int * packed_renumber_elements;
  int   length_of_packed_renumber_elements;
  int * renumber_index;
  int * packed_renumber_index;
  int * inv_packed_renumber_index;
  int * local_to_global;
  int   length_of_packed_renumber_index;
  int * index_reference;
  int * inv_packed_renumber_elements;

int length_of_packed_index_reference;
int * renum;
   double time_start;
   double time_end;


  printf("PE%i:\t renumber_sparse_matrix_kernel: \n",my_rank);
  printf("PE%i:\t    mat_number_of_rows = %d \n",my_rank,mat_number_of_rows);
  printf("PE%i:\t mat_number_of_entries = %d \n",my_rank,mat_number_of_entries);

   time_start=MPI_Wtime();
  renumber_elements =malloc(sizeof(int)*mat_number_of_rows);     //OK
  sort_int_up(mat_elements,renumber_elements,mat_number_of_rows);
   time_end=MPI_Wtime();
   printf("PE%i:\t renumber_sparse_matrix_kernel/sort_int_up(mat_elements,: time %f muesec \n",my_rank, (time_end-time_start)*1.e6);

//  print_indirect( "renumbered elements", renumber_elements, mat_elements, mat_number_of_rows);


/* We assume that non unique mat_elements[] are allowed.
 * If mat_elements[] are unique the following step can be simplified.
 */
   inv_packed_renumber_elements   =malloc(sizeof(int)*mat_number_of_rows);  //OK
   index_reference    =malloc(sizeof(int)*mat_number_of_entries);           //OK
   packed_renumber_elements    =malloc(sizeof(int)*mat_number_of_rows);     //OK
   length_of_packed_renumber_elements=pack_renumbering_index(mat_elements,renumber_elements,mat_number_of_rows
                                                            ,packed_renumber_elements,inv_packed_renumber_elements); 

 // print_indirect( "packed renumbered elements", packed_renumber_elements, mat_elements, length_of_packed_renumber_elements);
  //print_aligned( "mapping of elements", mat_elements, inv_packed_renumber_elements, mat_number_of_rows);


print_array("mat_index: ",mat_index,mat_number_of_entries);

   time_start=MPI_Wtime();
  renumber_index    =malloc(sizeof(int)*mat_number_of_entries);     //OK
  sort_int_up(mat_index,renumber_index,mat_number_of_entries);
   time_end=MPI_Wtime();
   printf("PE%i:\t renumber_sparse_matrix_kernel/sort_int_up(mat_index,: time %f muesec \n",my_rank, (time_end-time_start)*1.e6);

  print_indirect( "renumbered indices", renumber_index, mat_index, mat_number_of_entries);


   time_start=MPI_Wtime();
    length_local_to_global=identify_global_indices_in_global_elements(
                                                mat_elements,packed_renumber_elements,length_of_packed_renumber_elements
                                               ,mat_index   ,renumber_index,mat_number_of_entries
                                               ,index_reference,mat_number_of_entries   
                                               );
   time_end=MPI_Wtime();
   printf("PE%i:\t renumber_sparse_matrix_kernel/identify_global_indices_in_global_elements: time %f muesec \n",my_rank, (time_end-time_start)*1.e6);

/*
 * mat_elements[packed_renumber_elements[]] maps new elements to global elements
 * index_reference[] maps origin of the mat_indices to new values
 */

print_array("index_reference: ",index_reference,mat_number_of_entries);

	local_to_global    =malloc(sizeof(int)*length_local_to_global); // geht zurueck
   for(ii=0;ii<mat_number_of_entries; ii++) {
//	local_to_global[index_reference[ii]]=mat_index[index_reference[ii]];
	local_to_global[index_reference[ii]]=mat_index[ii];
 if (index_reference[ii] <  0 || index_reference[ii]>= length_local_to_global) {
    printf("PE%i:\t error :\n",my_rank);
    printf("PE%i:\t ii=%d,index_reference[ii]=%d,length_local_to_global= %d \n",my_rank,
                    ii   ,index_reference[ii]   ,length_local_to_global);
     exit_MPI(1);
    }
	}

/* now all global indizes are defined
 * still to define the element values
 */
   for(jj=0;jj<length_of_packed_renumber_elements; jj++) {
	local_to_global[jj]=mat_elements[packed_renumber_elements[jj]];
	}
/*
 * length_of_packed_renumber_elements is the length of the inner indices part
 */
   loc_to_global->renumber=local_to_global;
   loc_to_global->inner_length=length_of_packed_renumber_elements;
   loc_to_global->total_length=length_local_to_global;

  //print_indirect("   no   local global        elements ", packed_renumber_elements, mat_elements, length_of_packed_renumber_elements);
  //print_aligned("   no   local global        indices ", index_reference, mat_index, mat_number_of_entries);
  printf("PE%i:\t length_local_to_global=%d\n",my_rank,length_local_to_global);


///* only for testing begin */
//   renum=malloc(sizeof(int)*mat_number_of_entries);   // OK
//   packed_renumber_index=malloc(sizeof(int)*mat_number_of_entries);   // OK
//   inv_packed_renumber_index=malloc(sizeof(int)*mat_number_of_entries);   // OK
//
//    sort_int_up(index_reference,renum,mat_number_of_entries);
//   length_of_packed_index_reference=pack_renumbering_index(index_reference,renum,mat_number_of_entries
//                                                         ,packed_renumber_index,inv_packed_renumber_index); 
//
//
//   printf("PE%i:\t length_of_packed_index_reference=%d\n",my_rank,length_of_packed_index_reference);
//
//  print_indirect( "packed renumbered index_reference ", packed_renumber_index, index_reference, length_of_packed_index_reference);
//
//  free(renum);
//  free(packed_renumber_index);
///* only for testing end */

  for(jj=0; jj < mat_number_of_rows; jj++){
		// printf("PE%i:\thallo %d %d \n",my_rank,mat_elements[jj],inv_packed_renumber_elements[jj]);
	mat_elements[jj]=inv_packed_renumber_elements[jj];
	}
  for(ii=0; ii < mat_number_of_entries; ii++){
		// printf("PE%i:\thallo %d %d \n",my_rank,mat_elements[jj],inv_packed_renumber_elements[jj]);
	mat_index[ii]=index_reference[ii];
	}

  free(renumber_index);
  free(renumber_elements);
  free(index_reference);
  free(packed_renumber_elements);
  free(inv_packed_renumber_elements);
/*
  fflush(NULL);
  exit_MPI(9);
  exit(9);
*/

}

/**************************************************************************************/
void alloc_sparse_matrix(sparse_matrix_type * matrix, int number_of_rows, int number_of_entries) {

 if (number_of_rows <  0 ) {
 printf("PE%i:\t error in routine alloc_sparse_matrix:\n",my_rank);
 printf("PE%i:\t number_of_rows= %d \n",my_rank,number_of_rows);
 printf("PE%i:\t is less than 0!\n",my_rank);
  exit_MPI(1);
  exit(1);
 }

 if ( number_of_entries <  0 ) {
 printf("PE%i:\t error in routine alloc_sparse_matrix:",my_rank);
 printf("PE%i:\t number_of_entries=%d \n",my_rank, number_of_entries);
 printf("PE%i:\t is less than 0!",my_rank);
  exit_MPI(1);
  exit(1);
 }

  if( number_of_rows > 0 && number_of_entries > 0 ) {
  matrix->matrix_type   = 0 ;
  matrix->mat_number_of_rows   = number_of_rows ;
  matrix->mat_number_of_entries= number_of_entries ;

  matrix->mat_number_of_row_elements=malloc(sizeof(int)   *number_of_rows);
  matrix->mat_index                 =malloc(sizeof(int)   *number_of_entries);
  matrix->mat_elements              =malloc(sizeof(int)   *number_of_rows);
  matrix->mat_aa                    =malloc(sizeof(double)*number_of_entries);

  } else {

  matrix->mat_number_of_rows   = 0 ;
  matrix->mat_number_of_entries= 0 ;

  matrix->mat_number_of_row_elements= NULL;
  matrix->mat_index                 = NULL;
  matrix->mat_elements              = NULL;
  matrix->mat_aa                    = NULL;

  }
}


/**************************************************************************************/
void dealloc_sparse_matrix(sparse_matrix_type * matrix) {

 if (matrix->mat_number_of_rows > 0 ) {
 free(matrix->mat_number_of_row_elements);
 free(matrix->mat_index                 );
 free(matrix->mat_elements              );
 free(matrix->mat_aa                    );
 }
   matrix->mat_number_of_rows=0;
   matrix->mat_number_of_entries=0;
}


/**************************************************************************************/
void set_simple_row_ordered_mat_val(sparse_matrix_type * mat_A, patch_type  patch,int jmx_glob) {
   double time_start;
   double time_end;

   time_start=MPI_Wtime();
     set_simple_row_ordered_mat_val_kernel_2( mat_A->mat_aa ,
                                            mat_A->mat_index ,
                                            mat_A->mat_elements ,
                                            mat_A->mat_number_of_row_elements ,
                                           &mat_A->mat_number_of_rows ,
                                           &mat_A->mat_number_of_entries ,
                                           &mat_A->matrix_type ,
                                           patch.patch_bounds, jmx_glob,patch.math_bounds
                                          );
   time_end=MPI_Wtime();
   printf("PE%i:\t set_simple_row_ordered_mat_val: time %f muesec \n",my_rank, (time_end-time_start)*1.e6);
           }
/**************************************************************************************/
void set_simple_row_ordered_mat_val_kernel(  
                                                  double * mat_aa  
                                                 ,int * mat_index  
                                                 ,int * mat_elements  
                                                 ,int * mat_number_of_row_elements  
                                                 ,int * mat_number_of_rows  
                                                 ,int * mat_number_of_entries  
                                                 ,int * matrix_type  
                                                 ,int * patch_bounds, int jmx_glob,int * math_bound)
{
      int ii, num, ind, jj, kk, ll ;
      int num_row_elements ;
      int jmin,jmax, kmin,kmax;
      int jmin_d,jmax_d, kmin_d,kmax_d;

/* Dirichlet Randbedingungen werden besetzt.
 * jmin,kmin und kmin,kmax geben die Eckindices des Teilgebietes an.
 * Sie sind global zu verstehen.
 * jmin,kmin muessen  >= 1 sein. 
 * Die Diskretisierung ist eine einfache 5-Punkt Diskretisierung mit quadratischen Maschen.
 * Besetzt wird eine 9 Punkt Stencil.
 * Dieser Stencil kann veraendert werden.
 * Die Diagonale der Matrix ist 1 !
 * Die Zuweisung der Elemente erfolgt lexikographisch.
 * math_bound[0:3]== 0 fuer innere Punkte.
 * math_bound[0]  == 1 fuer Rand auf unterer Seite.
 * math_bound[1]  == 1 fuer Rand auf rechter Seite.
 * math_bound[2]  == 1 fuer Rand auf oberer Seite.
 * math_bound[3]  == 1 fuer Rand auf linker Seite.


 * matrix is used to determine the halo.
 *
 */

   for(ll=0;ll<4;ll++) {
   if(math_bound[0] != 0 && math_bound[0] != 1) {
       printf("PE%i:\t error in routine set_simple_row_ordered_mat_val_kernel:",my_rank);
       printf("PE%i:\t math_bound[%d]=%d \n",my_rank,ll,math_bound[ll]);
       printf("PE%i:\t is nor 0 nor 1",my_rank);
       exit_MPI(1);
       exit(1);
      }
   }
   jmin=patch_bounds[0];
   kmin=patch_bounds[1];
   jmax=patch_bounds[2];
   kmax=patch_bounds[3];

   jmin_d=math_bound[3];
   jmax_d=math_bound[1];
   kmin_d=math_bound[0];
   kmax_d=math_bound[2];

      ii=-1;
      num=-1;
      num_row_elements=0;

if( math_bound[0] == 1 && math_bound[3] == 1 ) {
/* Ecke links unten   */
    kk=kmin;
    jj=jmin;
      ind=jj+jmx_glob*kk;
      ii=ii+1;

      mat_elements[ii]=ind;

      num_row_elements=-1;

      num=num+1; num_row_elements=num_row_elements+1; mat_aa[num]= 1.   ; mat_index[num]=ind;
      num=num+1; num_row_elements=num_row_elements+1; mat_aa[num]=-0.25 ; mat_index[num]=ind+1;
      num=num+1; num_row_elements=num_row_elements+1; mat_aa[num]=-0.25 ; mat_index[num]=ind+jmx_glob;
//null      num=num+1; num_row_elements=num_row_elements+1; mat_aa[num]= 0.   ; mat_index[num]=ind+jmx_glob+1;//no_null_elements
      mat_number_of_row_elements[ii]=num_row_elements+1;
}

if( math_bound[0] == 1 ) {
/* unterer Rand Mitte */
	printf("PE%i:\t hier unterer Rand Mitte \n",my_rank);
  kk=kmin;
   for( jj=jmin+jmin_d; jj<jmax-jmax_d+1;jj++ ) {
      ind=jj+jmx_glob*kk;
      ii=ii+1;

      mat_elements[ii]=ind;

      num_row_elements=-1;

      num=num+1; num_row_elements=num_row_elements+1; mat_aa[num]=-0.25 ; mat_index[num]=ind-1;
      num=num+1; num_row_elements=num_row_elements+1; mat_aa[num]= 1.   ; mat_index[num]=ind;
      num=num+1; num_row_elements=num_row_elements+1; mat_aa[num]=-0.25 ; mat_index[num]=ind+1;
//null      num=num+1; num_row_elements=num_row_elements+1; mat_aa[num]= 0.   ; mat_index[num]=ind+jmx_glob-1;//no_null_elements
      num=num+1; num_row_elements=num_row_elements+1; mat_aa[num]=-0.25 ; mat_index[num]=ind+jmx_glob;
//null      num=num+1; num_row_elements=num_row_elements+1; mat_aa[num]= 0.   ; mat_index[num]=ind+jmx_glob+1;//no_null_elements
      mat_number_of_row_elements[ii]=num_row_elements+1;
   }
}


if( math_bound[0] == 1 && math_bound[1] == 1 ) {
/* Ecke rechts unten  */
   kk=kmin;
   jj=jmax;
      ind=jj+jmx_glob*kk;
      ii=ii+1;

      mat_elements[ii]=ind;

      num_row_elements=-1;

      num=num+1; num_row_elements=num_row_elements+1; mat_aa[num]=-0.25 ; mat_index[num]=ind-1;
      num=num+1; num_row_elements=num_row_elements+1; mat_aa[num]= 1.   ; mat_index[num]=ind;
//null      num=num+1; num_row_elements=num_row_elements+1; mat_aa[num]= 0.   ; mat_index[num]=ind+jmx_glob-1;//no_null_elements
      num=num+1; num_row_elements=num_row_elements+1; mat_aa[num]=-0.25 ; mat_index[num]=ind+jmx_glob;
      mat_number_of_row_elements[ii]=num_row_elements+1;
}

   for (kk=kmin+kmin_d; kk<kmax-kmax_d+1; kk++){

if( math_bound[3] == 1 ) {
/* linker Rand Mitte   */
      jj=jmin ;
      ind=jj+jmx_glob*kk;
      ii=ii+1;

      mat_elements[ii]=ind;

      num_row_elements=-1;

      num=num+1; num_row_elements=num_row_elements+1; mat_aa[num]=-0.25 ; mat_index[num]=ind-jmx_glob;
//null      num=num+1; num_row_elements=num_row_elements+1; mat_aa[num]= 0.   ; mat_index[num]=ind-jmx_glob+1;//no_null_elements
      num=num+1; num_row_elements=num_row_elements+1; mat_aa[num]= 1.   ; mat_index[num]=ind;
      num=num+1; num_row_elements=num_row_elements+1; mat_aa[num]=-0.25 ; mat_index[num]=ind+1;
      num=num+1; num_row_elements=num_row_elements+1; mat_aa[num]=-0.25 ; mat_index[num]=ind+jmx_glob;
//null      num=num+1; num_row_elements=num_row_elements+1; mat_aa[num]= 0.   ; mat_index[num]=ind+jmx_glob+1;//no_null_elements
      mat_number_of_row_elements[ii]=num_row_elements+1;
}


/* inneres Teilgebiet */
    for( jj=jmin+jmin_d; jj<jmax-jmax_d+1;jj++ ) {
      ind=jj+jmx_glob*kk;
      ii=ii+1;

      mat_elements[ii]=ind;

      num_row_elements=-1;

//null      num=num+1; num_row_elements=num_row_elements+1; mat_aa[num]= 0.   ; mat_index[num]=ind-jmx_glob-1;//no_null_elements
      num=num+1; num_row_elements=num_row_elements+1; mat_aa[num]=-0.25 ; mat_index[num]=ind-jmx_glob;
//null      num=num+1; num_row_elements=num_row_elements+1; mat_aa[num]= 0.   ; mat_index[num]=ind-jmx_glob+1;//no_null_elements
      num=num+1; num_row_elements=num_row_elements+1; mat_aa[num]=-0.25 ; mat_index[num]=ind-1  ;
      num=num+1; num_row_elements=num_row_elements+1; mat_aa[num]= 1.   ; mat_index[num]=ind     ;
      num=num+1; num_row_elements=num_row_elements+1; mat_aa[num]=-0.25 ; mat_index[num]=ind+1 ;
//null      num=num+1; num_row_elements=num_row_elements+1; mat_aa[num]= 0.   ; mat_index[num]=ind+jmx_glob-1;//no_null_elements
      num=num+1; num_row_elements=num_row_elements+1; mat_aa[num]=-0.25 ; mat_index[num]=ind+jmx_glob;
//null      num=num+1; num_row_elements=num_row_elements+1; mat_aa[num]= 0.   ; mat_index[num]=ind+jmx_glob+1;//no_null_elements

      mat_number_of_row_elements[ii]=num_row_elements+1;

    }


if( math_bound[1] == 1 ) {
/* rechter Rand Mitte  */
   jj=jmax ;
      ind=jj+jmx_glob*kk;
      ii=ii+1;

      mat_elements[ii]=ind;

      num_row_elements=-1;

//null      num=num+1; num_row_elements=num_row_elements+1; mat_aa[num]= 0.   ; mat_index[num]=ind-jmx_glob-1;//no_null_elements
      num=num+1; num_row_elements=num_row_elements+1; mat_aa[num]=-0.25 ; mat_index[num]=ind-jmx_glob;
      num=num+1; num_row_elements=num_row_elements+1; mat_aa[num]=-0.25 ; mat_index[num]=ind-1;
      num=num+1; num_row_elements=num_row_elements+1; mat_aa[num]= 1.   ; mat_index[num]=ind;
//null      num=num+1; num_row_elements=num_row_elements+1; mat_aa[num]= 0.   ; mat_index[num]=ind+jmx_glob-1;//no_null_elements
      num=num+1; num_row_elements=num_row_elements+1; mat_aa[num]=-0.25 ; mat_index[num]=ind+jmx_glob;
      mat_number_of_row_elements[ii]=num_row_elements+1;

}
}


if( math_bound[2] == 1 && math_bound[3] == 1 ) {
/* Ecke links oben  */
   kk=kmax;
   jj=jmin;
      ind=jj+jmx_glob*kk;
      ii=ii+1;

      mat_elements[ii]=ind;

      num_row_elements=-1;

      num=num+1; num_row_elements=num_row_elements+1; mat_aa[num]=-0.25 ; mat_index[num]=ind-jmx_glob;
//null      num=num+1; num_row_elements=num_row_elements+1; mat_aa[num]=-0.00 ; mat_index[num]=ind-jmx_glob+1;//no_null_elements
      num=num+1; num_row_elements=num_row_elements+1; mat_aa[num]= 1.   ; mat_index[num]=ind;
      num=num+1; num_row_elements=num_row_elements+1; mat_aa[num]=-0.25 ; mat_index[num]=ind+1;
      mat_number_of_row_elements[ii]=num_row_elements+1;
}


if( math_bound[2] == 1 ) {
/* oberer Rand  Mitte */
   kk=kmax;
   for( jj=jmin+jmin_d; jj<jmax-jmax_d+1;jj++ ) {
      ind=jj+jmx_glob*kk;
      ii=ii+1;

      mat_elements[ii]=ind;

      num_row_elements=-1;

//null      num=num+1; num_row_elements=num_row_elements+1; mat_aa[num]=-0.00 ; mat_index[num]=ind-jmx_glob-1;//no_null_elements
      num=num+1; num_row_elements=num_row_elements+1; mat_aa[num]=-0.25 ; mat_index[num]=ind-jmx_glob;
//null      num=num+1; num_row_elements=num_row_elements+1; mat_aa[num]=-0.00 ; mat_index[num]=ind-jmx_glob+1;//no_null_elements
      num=num+1; num_row_elements=num_row_elements+1; mat_aa[num]=-0.25 ; mat_index[num]=ind-1;
      num=num+1; num_row_elements=num_row_elements+1; mat_aa[num]= 1.   ; mat_index[num]=ind;
      num=num+1; num_row_elements=num_row_elements+1; mat_aa[num]=-0.25 ; mat_index[num]=ind+1;
      mat_number_of_row_elements[ii]=num_row_elements+1;
   }
}


if( math_bound[1] == 1 && math_bound[2] == 1 ) {
/* Ecke rechts oben */
   kk=kmax;
   jj=jmax;
      ind=jj+jmx_glob*kk;
      ii=ii+1;

      mat_elements[ii]=ind;

      num_row_elements=-1;

//null      num=num+1; num_row_elements=num_row_elements+1; mat_aa[num]=-0.00 ; mat_index[num]=ind-jmx_glob-1;//no_null_elements
      num=num+1; num_row_elements=num_row_elements+1; mat_aa[num]=-0.25 ; mat_index[num]=ind-jmx_glob;
      num=num+1; num_row_elements=num_row_elements+1; mat_aa[num]=-0.25 ; mat_index[num]=ind-1;
      num=num+1; num_row_elements=num_row_elements+1; mat_aa[num]= 1.   ; mat_index[num]=ind;
      mat_number_of_row_elements[ii]=num_row_elements+1;
}

/* Besetzung der tatsaechlichen Parameter   */

      *mat_number_of_rows=ii+1;
      *mat_number_of_entries=num+1;
      *matrix_type=0;

      }
/**************************************************************************************/
void set_simple_row_ordered_mat_val_kernel_2(  
                                                  double * mat_aa  
                                                 ,int * mat_index  
                                                 ,int * mat_elements  
                                                 ,int * mat_number_of_row_elements  
                                                 ,int * mat_number_of_rows  
                                                 ,int * mat_number_of_entries  
                                                 ,int * matrix_type  
                                                 ,int * patch_bounds, int jmx_glob,int * math_bound)
{
      int ii, num, ind, jj, kk, ll ;
      int num_row_elements ;
      int jmin,jmax, kmin,kmax;
      int jmin_d,jmax_d, kmin_d,kmax_d;

/* Dirichlet Randbedingungen werden besetzt.
 * jmin,kmin und kmin,kmax geben die Eckindices des Teilgebietes an.
 * Sie sind global zu verstehen.
 * jmin,kmin muessen  >= 1 sein. 
 * Die Diskretisierung ist eine einfache 5-Punkt Diskretisierung mit quadratischen Maschen.
 * Besetzt wird eine 9 Punkt Stencil.
 * Dieser Stencil kann veraendert werden.
 * Die Diagonale der Matrix ist 1 !
 * Die Zuweisung der Elemente erfolgt lexikographisch.
 * math_bound[0:3]== 0 fuer innere Punkte.
 * math_bound[0]  == 1 fuer Rand auf unterer Seite.
 * math_bound[1]  == 1 fuer Rand auf rechter Seite.
 * math_bound[2]  == 1 fuer Rand auf oberer Seite.
 * math_bound[3]  == 1 fuer Rand auf linker Seite.


 * matrix is used to determine the halo.
 *
 */

   for(ll=0;ll<4;ll++) {
   if(math_bound[0] != 0 && math_bound[0] != 1) {
       printf("PE%i:\t error in routine set_simple_row_ordered_mat_val_kernel:",my_rank);
       printf("PE%i:\t math_bound[%d]=%d \n",my_rank,ll,math_bound[ll]);
       printf("PE%i:\t is nor 0 nor 1",my_rank);
       exit_MPI(1);
       exit(1);
      }
   }
   jmin=patch_bounds[0];
   kmin=patch_bounds[1];
   jmax=patch_bounds[2];
   kmax=patch_bounds[3];

   jmin_d=math_bound[3];
   jmax_d=math_bound[1];
   kmin_d=math_bound[0];
   kmax_d=math_bound[2];

      ii=-1;
      num=-1;
      num_row_elements=0;


// The matrix contains all equations including the near boundary equations.
// The boundary terms involved in the near boundary equations have to be extracted
// by a special procedure.


//   for (kk=kmin+kmin_d; kk<kmax-kmax_d+1; kk++){
//
//    for( jj=jmin+jmin_d; jj<jmax-jmax_d+1;jj++ ) {

   for (kk=kmin; kk<kmax+1; kk++){

    for( jj=jmin; jj<jmax+1;jj++ ) {
      ind=jj+jmx_glob*kk;
      ii=ii+1;

      mat_elements[ii]=ind;

      num_row_elements=-1;

//null      num=num+1; num_row_elements=num_row_elements+1; mat_aa[num]= 0.   ; mat_index[num]=ind-jmx_glob-1;//no_null_elements
      num=num+1; num_row_elements=num_row_elements+1; mat_aa[num]=-0.25 ; mat_index[num]=ind-jmx_glob;
//null      num=num+1; num_row_elements=num_row_elements+1; mat_aa[num]= 0.   ; mat_index[num]=ind-jmx_glob+1;//no_null_elements
      num=num+1; num_row_elements=num_row_elements+1; mat_aa[num]=-0.25 ; mat_index[num]=ind-1  ;
      num=num+1; num_row_elements=num_row_elements+1; mat_aa[num]= 1.   ; mat_index[num]=ind     ;
      num=num+1; num_row_elements=num_row_elements+1; mat_aa[num]=-0.25 ; mat_index[num]=ind+1 ;
//null      num=num+1; num_row_elements=num_row_elements+1; mat_aa[num]= 0.   ; mat_index[num]=ind+jmx_glob-1;//no_null_elements
      num=num+1; num_row_elements=num_row_elements+1; mat_aa[num]=-0.25 ; mat_index[num]=ind+jmx_glob;
//null      num=num+1; num_row_elements=num_row_elements+1; mat_aa[num]= 0.   ; mat_index[num]=ind+jmx_glob+1;//no_null_elements

      mat_number_of_row_elements[ii]=num_row_elements+1;

    }


}



/* Besetzung der tatsaechlichen Parameter   */

      *mat_number_of_rows=ii+1;
      *mat_number_of_entries=num+1;
      *matrix_type=0;

      }
/**************************************************************************************/
   void decompose_matrix( sparse_matrix_type matrix,sparse_matrix_type * matrix_new,int * marker)
 {
       int number_of_rows, number_of_entries;
   double time_start;
   double time_end;

   time_start=MPI_Wtime();

           decompose_matrix_kernel_size(
                                   matrix.mat_index
                                  ,matrix.mat_elements
                                  ,matrix.mat_number_of_row_elements
                                  ,matrix.mat_number_of_rows
                                  ,matrix.mat_number_of_entries
                                  ,&number_of_rows
                                  ,&number_of_entries
                                  ,marker
                                  ) ;

           alloc_sparse_matrix(matrix_new, number_of_rows, number_of_entries) ;

           decompose_matrix_kernel_data(
                                   matrix.mat_aa
                                  ,matrix.mat_index
                                  ,matrix.mat_elements
                                  ,matrix.mat_number_of_row_elements
                                  ,matrix.mat_number_of_rows
                                  ,matrix.mat_number_of_entries
                                  ,matrix_new->mat_aa
                                  ,matrix_new->mat_index
                                  ,matrix_new->mat_elements
                                  ,matrix_new->mat_number_of_row_elements
                                  ,&(matrix_new->mat_number_of_rows)
                                  ,&(matrix_new->mat_number_of_entries)
                                  ,&(matrix_new->matrix_type)
                                  ,marker
                                  ) ;
   time_end=MPI_Wtime();
   printf("PE%i:\t decompose_matrix: time %f muesec \n",my_rank, (time_end-time_start)*1.e6);
 }
/**************************************************************************************/
   void decompose_matrix_kernel_size(
                                   int * mat_index
                                  ,int * mat_elements
                                  ,int * mat_number_of_row_elements
                                  ,int mat_number_of_rows
                                  ,int mat_number_of_entries
                                  ,int * mat_number_of_rows_new
                                  ,int * mat_number_of_entries_new
                                  ,int * marker
                                  )
{
 int ii,nn,kk,num,nn_max;
 int ii_new, num_new, no_row_ind_new;


    *mat_number_of_rows_new=0;
    *mat_number_of_entries_new=0;

 num=-1;
 ii_new=0;
 num_new=0;
 for( ii = 0 ; ii < mat_number_of_rows; ii++ ) {
      nn_max = mat_number_of_row_elements[ii];
      no_row_ind_new=0;
     for( nn = 0; nn < nn_max ; nn++) {
          num = num+1;
      if( marker[mat_index[num]] == 1 ){
           no_row_ind_new=no_row_ind_new+1;
                      num_new=num_new+1;
       }
     }

   if( no_row_ind_new > 0 ){
     ii_new=ii_new+1;
   }
 }
    *mat_number_of_rows_new=ii_new;
    *mat_number_of_entries_new=num_new;

}
/**************************************************************************************/
   void decompose_matrix_kernel_data(
                                   double * mat_aa
                                  ,int * mat_index
                                  ,int * mat_elements
                                  ,int * mat_number_of_row_elements
                                  ,int mat_number_of_rows
                                  ,int mat_number_of_entries
                                  ,double * mat_aa_new
                                  ,int * mat_index_new
                                  ,int * mat_elements_new
                                  ,int * mat_number_of_row_elements_new
                                  ,int * mat_number_of_rows_new
                                  ,int * mat_number_of_entries_new
                                  ,int * matrix_type  
                                  ,int * marker
                                  )
{
 int ii,nn,kk,num,nn_max;
 int ii_new, num_new, no_row_ind_new;


    *mat_number_of_rows_new=0;
    *mat_number_of_entries_new=0;

 num=-1;
 ii_new=0;
 num_new=0;
 for( ii = 0 ; ii < mat_number_of_rows; ii++ ) {
      nn_max = mat_number_of_row_elements[ii];
      no_row_ind_new=0;
     for( nn = 0; nn < nn_max ; nn++) {
          num = num+1;
      if( marker[mat_index[num]] == 1 ){
           no_row_ind_new=no_row_ind_new+1;

          mat_aa_new[num_new]=  -mat_aa[num];   // remark the switched sign!
       mat_index_new[num_new]=mat_index[num];   
                      num_new=num_new+1;

                  mat_aa[num]=0.;
       }
     }

   if( no_row_ind_new > 0 ){
     mat_number_of_row_elements_new[ii_new]=no_row_ind_new;
     mat_elements_new[ii_new]=mat_elements[ii];
     ii_new=ii_new+1;
   }
 }
    *mat_number_of_rows_new=ii_new;
    *mat_number_of_entries_new=num_new;
    *matrix_type=0;

}
/**************************************************************************************/
